/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.test.messaging.core.plugin.base;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.jboss.messaging.core.Message;
import org.jboss.messaging.core.MessageReference;
import org.jboss.test.messaging.MessagingTestCase;
import org.jboss.test.messaging.tools.jmx.ServiceContainer;
import org.jboss.messaging.core.message.MessageFactory;
import org.jboss.messaging.core.plugin.contract.MessageStore;


/**
 * The test strategy is to try as many combination as it makes sense of the following
 * variables:
 *
 * 1. The message store can be can be non-recoverable (does not have access to a
 *    PersistenceManager) or recoverable. A non-recoverable message store can accept reliable
 *    messages or not.
 * 2. The message can be non-reliable or reliable.
 * 3. One or multiple messages can be stored.
 * 4. One or more strong references may be maintain for the same MessageReference.
 * 5. A recoverable message store may be forcibly crashed and tested if it recovers.
 *
 * @author <a href="mailto:ovidiu@jboss.org">Ovidiu Feodorov</a>
 * @version <tt>$Revision: 764 $</tt>
 *
 * $Id: MessageStoreTestBase.java 764 2006-03-22 10:23:38Z timfox $
 */
public abstract class MessageStoreTestBase extends MessagingTestCase
{
   // Constants -----------------------------------------------------

   public static final int NUMBER_OF_MESSAGES = 100;

   // Static --------------------------------------------------------       

   /**
    * Makes sure the given reference refers the given message.
    */
   public static void assertCorrectReference(MessageReference ref, Object storeID, Message m)
   {
      assertEquals(storeID, ref.getStoreID());
      assertTrue(ref.isReference());
      assertEquals(m.getMessageID(), ref.getMessageID());
      assertEquals(m.isReliable(), ref.isReliable());
      assertEquals(m.getExpiration(), ref.getExpiration());
      assertEquals(m.getTimestamp(), ref.getTimestamp());
      assertEquals(m.getPriority(), ref.getPriority());
      assertFalse(ref.isRedelivered());

      Map messageHeaders = m.getHeaders();
      Map refHeaders = ref.getHeaders();
      assertEquals(messageHeaders.size(), refHeaders.size());

      for(Iterator i = messageHeaders.keySet().iterator(); i.hasNext(); )
      {
         Object header = i.next();
         Object value = messageHeaders.get(header);
         assertTrue(refHeaders.containsKey(header));
         assertEquals(value, refHeaders.get(header));
      }
   }

   // Attributes ----------------------------------------------------

   protected ServiceContainer sc;
   protected MessageStore ms;

   protected Map headers;

   // Constructors --------------------------------------------------

   public MessageStoreTestBase(String name)
   {
      super(name);
   }

   // Public --------------------------------------------------------

   public void setUp() throws Exception
   {
      super.setUp();

      sc = new ServiceContainer("all,-remoting,-security");
      sc.start();

      headers = new HashMap();
      headers.put("someKey", "someValue");
   }

   public void tearDown() throws Exception
   {
      headers.clear();
      headers = null;
      sc.stop();
      sc = null;

      super.tearDown();
   }

   //
   // Non-recoverable message store
   //

   ////
   //// Non-reliable message
   ////

   //////
   ////// One message
   //////

   ////////
   //////// One strong reference
   ////////

   public void testNonRecoverableMessageStore_1() throws Exception
   {
      if (ms.isRecoverable())
      {
         // we only test non-recoverable message stores
         return;
      }

      Message m =
         MessageFactory.createCoreMessage(0, false, 777l, 888l, (byte)9, headers, "payload");

      // non-recoverable store, non-reliable message, one message
      MessageReference ref = ms.reference(m);
      assertCorrectReference(ref, ms.getStoreID(), m);

      assertCorrectReference(ref, ms.getStoreID(), m);
  
      ref.releaseMemoryReference();

      // the message store should be cleaned of everything that pertains to that message

      ref = ms.reference(m.getMessageID());
      assertNull(ref);
   }

   ////////
   //////// Two strong references
   ////////

   public void testNonRecoverableMessageStore_1_1() throws Exception
   {
      if (ms.isRecoverable())
      {
         // we only test non-recoverable message stores
         return;
      }

      Message m =
         MessageFactory.createCoreMessage(0, false, 777l, 888l, (byte)9, headers, "payload");

      // non-recoverable store, non-reliable message, one message
      MessageReference ref = ms.reference(m);
      assertCorrectReference(ref, ms.getStoreID(), m);

      MessageReference ref2 = ms.reference(m);

      assertCorrectReference(ref2, ms.getStoreID(), m);
   }

   //////
   ////// Multiple messages
   //////

   public void testNonRecoverableMessageStore_2() throws Exception
   {
      if (ms.isRecoverable())
      {
         // we only test non-recoverable message stores
         return;
      }

      Message[] m = new Message[NUMBER_OF_MESSAGES];
      MessageReference[] refs = new MessageReference[NUMBER_OF_MESSAGES];
      for(int i = 0; i < NUMBER_OF_MESSAGES; i++ )
      {
         m[i] = MessageFactory.createCoreMessage(i, false, 700 + i, 800 + i,
                                             (byte)(i % 10), headers, "payload" + i);

         // non-recoverable store, non-reliable message, one message
         refs[i] = ms.reference(m[i]);
         assertCorrectReference(refs[i], ms.getStoreID(), m[i]);
      }

      for (int i = 0; i < NUMBER_OF_MESSAGES; i++)
      {         
         assertCorrectReference(refs[i], ms.getStoreID(), m[i]);
         refs[i].releaseMemoryReference();
      }

      // there are no strong references to the message reference anymore, so the message store
      // should be cleaned of everything that pertains to those message

      for (int i = 0; i < NUMBER_OF_MESSAGES; i++)
      {
         assertNull(ms.reference(m[i].getMessageID()));
      }
   }

   ////
   //// Reliable message
   ////

   //////
   ////// The non-reliable store does NOT accept reliable messages
   //////

   //////
   ////// One message
   //////

   public void testNonRecoverableMessageStore_3() throws Exception
   {
      if (ms.isRecoverable())
      {
         // we only test non-recoverable message stores
         return;
      }

      if (ms.acceptReliableMessages())
      {
         // we only test message stores that do not accept reliable message
         return;
      }

      Message m = MessageFactory.
      createCoreMessage(0, true, 777l, 888l, (byte)9, headers, "payload");

      // non-recoverable store, reliable message, one message
      try
      {
         ms.reference(m);
         fail("should throw IllegalStateException");
      }
      catch(IllegalStateException e)
      {
         // OK
      }
   }

   //////
   ////// The non-reliable store accepts reliable messages
   //////

   ////////
   //////// One message
   ////////

   public void testNonRecoverableMessageStore_4() throws Exception
   {
      if (ms.isRecoverable())
      {
         // we only test non-recoverable message stores
         return;
      }

      if (!ms.acceptReliableMessages())
      {
         // we only test message stores that accept reliable message
         return;
      }

      Message m = MessageFactory.
      createCoreMessage(0, false, 777l, 888l, (byte)9, headers, "payload");

      // non-recoverable store, non-reliable message, one message
      MessageReference ref = ms.reference(m);
      assertCorrectReference(ref, ms.getStoreID(), m);
      
      ref.releaseMemoryReference();

      // there's no strong reference to the unique message reference anymore, so the message store
      // should be cleaned of everything that pertains to that message

      ref = ms.reference(m.getMessageID());
      assertNull(ref);
   }

   ////////
   //////// Multiple messages
   ////////

   public void testNonRecoverableMessageStore_5() throws Exception
   {
      if (ms.isRecoverable())
      {
         // we only test non-recoverable message stores
         return;
      }

      if (!ms.acceptReliableMessages())
      {
         // we only test message stores that accept reliable message
         return;
      }

      Message[] m = new Message[NUMBER_OF_MESSAGES];
      MessageReference[] refs = new MessageReference[NUMBER_OF_MESSAGES];
      for(int i = 0; i < NUMBER_OF_MESSAGES; i++ )
      {
         m[i] = MessageFactory.createCoreMessage(i, false, 700 + i, 800 + i,
                                             (byte)(i % 10), headers, "payload" + i);

         // non-recoverable store, non-reliable message, one message
         refs[i] = ms.reference(m[i]);
         assertCorrectReference(refs[i], ms.getStoreID(), m[i]);
      }

      // get reachable reference

      for (int i = 0; i < NUMBER_OF_MESSAGES; i++)
      {         
         assertCorrectReference(refs[i], ms.getStoreID(), m[i]);
         refs[i].releaseMemoryReference();
      }

      // there are no strong references to the message reference anymore, so the message store
      // should be cleaned of everything that pertains to those message

      for (int i = 0; i < NUMBER_OF_MESSAGES; i++)
      {
         assertNull(ms.reference(m[i].getMessageID()));
      }
   }
   
   public void testNonRecoverableMessageStore_6() throws Exception
   {
      if (ms.isRecoverable())
      {
         // we only test non-recoverable message stores
         return;
      }

      if (!ms.acceptReliableMessages())
      {
         // we only test message stores that accept reliable message
         return;
      }

      Message m = MessageFactory.
      createCoreMessage(0, false, 777l, 888l, (byte)9, headers, "payload");

      // non-recoverable store, non-reliable message, one message
      MessageReference ref = ms.reference(m);
      assertCorrectReference(ref, ms.getStoreID(), m);
      
      //Create another reference from that reference
      MessageReference ref2 = ms.reference(m.getMessageID());
      
      ref.releaseMemoryReference();
      
      MessageReference ref3 = ms.reference(m.getMessageID());
      assertNotNull(ref3);
      
      assertCorrectReference(ref3, ms.getStoreID(), m);
      
      ref2.releaseMemoryReference();
      
      ref3.releaseMemoryReference();

      // there's no strong reference to the unique message reference anymore, so the message store
      // should be cleaned of everything that pertains to that message

      ref3 = ms.reference(m.getMessageID());
      assertNull(ref3);
   }
   
   public void testNonRecoverableMessageStore_7() throws Exception
   {
      if (ms.isRecoverable())
      {
         // we only test non-recoverable message stores
         return;
      }

      if (!ms.acceptReliableMessages())
      {
         // we only test message stores that accept reliable message
         return;
      }

      Message m = MessageFactory.
      createCoreMessage(0, false, 777l, 888l, (byte)9, headers, "payload");

      // non-recoverable store, non-reliable message, one message
      MessageReference ref = ms.reference(m);
      assertCorrectReference(ref, ms.getStoreID(), m);
      
      MessageReference ref2 = ms.reference(ref.getMessageID());
      assertNotNull(ref2);      
      assertCorrectReference(ref2, ms.getStoreID(), m);
        
      MessageReference ref3 = ms.reference(ref.getMessageID());
      assertNotNull(ref3);      
      assertCorrectReference(ref3, ms.getStoreID(), m);
      
      ref.releaseMemoryReference();
      ref2.releaseMemoryReference();
      ref3.releaseMemoryReference();
      
      MessageReference ref4 = ms.reference(ref.getMessageID());
 
      assertNull(ref4);    
      
   }

   //
   // Recoverable message store
   //

   ////
   //// Non-reliable message
   ////

   //////
   ////// One message
   //////

   ////////
   //////// One strong reference
   ////////

   public void testRecoverableMessageStore_1() throws Exception
   {
      if (!ms.isRecoverable())
      {
         // we only test recoverable message stores
         return;
      }

      Message m = MessageFactory.
      createCoreMessage(0, false, 777l, 888l, (byte)9, headers, "payload");

      // recoverable store, non-reliable message, one message
      MessageReference ref = ms.reference(m);
      assertCorrectReference(ref, ms.getStoreID(), m);
      ref.releaseMemoryReference();

      ref = ms.reference(m.getMessageID());
      assertNull(ref);
   }

   ////////
   //////// Two strong references
   ////////

   public void testRecoverableMessageStore_1_1() throws Exception
   {
      if (!ms.isRecoverable())
      {
         // we only test recoverable message stores
         return;
      }

      Message m = MessageFactory.
      createCoreMessage(0, false, 777l, 888l, (byte)9, headers, "payload");

      // recoverable store, non-reliable message, one message
      MessageReference ref = ms.reference(m);
      assertCorrectReference(ref, ms.getStoreID(), m);

      MessageReference ref2 = ms.reference(m);

      assertCorrectReference(ref2, ms.getStoreID(), m);
   }

   //////
   ////// Multiple messages
   //////

   public void testRecoverableMessageStore_2() throws Exception
   {
      if (!ms.isRecoverable())
      {
         // we only test recoverable message stores
         return;
      }

      Message[] m = new Message[NUMBER_OF_MESSAGES];
      MessageReference[] refs = new MessageReference[NUMBER_OF_MESSAGES];
      for(int i = 0; i < NUMBER_OF_MESSAGES; i++ )
      {
         m[i] = MessageFactory.createCoreMessage(i, false, 700 + i, 800 + i,
                                             (byte)(i % 10), headers, "payload" + i);

         // recoverable store, non-reliable message, one message
         refs[i] = ms.reference(m[i]);
         assertCorrectReference(refs[i], ms.getStoreID(), m[i]);
      }

      // get reachable reference

      for (int i = 0; i < NUMBER_OF_MESSAGES; i++)
      {         
         assertCorrectReference(refs[i], ms.getStoreID(), m[i]);
         refs[i].releaseMemoryReference();
      }

      for (int i = 0; i < NUMBER_OF_MESSAGES; i++)
      {
         assertNull(ms.reference(m[i].getMessageID()));
      }
   }

   ////
   //// Reliable message
   ////

   //////
   ////// One message
   //////

   ////////
   //////// One strong reference
   ////////

   public void testRecoverableMessageStore_3() throws Exception
   {
      if (!ms.isRecoverable())
      {
         // we only test recoverable message stores
         return;
      }

      Message m = MessageFactory.
      createCoreMessage(0, true, 777l, 888l, (byte)9, headers, "payload");

      // recoverable store, non-reliable message, one message
      MessageReference ref = ms.reference(m);
      assertCorrectReference(ref, ms.getStoreID(), m);

      ref.releaseMemoryReference();

      // there's no strong reference to the unique message reference anymore, so the message store
      // should be cleaned of everything that pertains to that message

      ref = ms.reference(m.getMessageID());
      assertNull(ref);
   }

   ////////
   //////// Two strong references
   ////////

   public void testRecoverableMessageStore_3_1() throws Exception
   {
      if (!ms.isRecoverable())
      {
         // we only test recoverable message stores
         return;
      }

      Message m = MessageFactory.
      createCoreMessage(0, true, 777l, 888l, (byte)9, headers, "payload");

      // recoverable store, non-reliable message, one message
      MessageReference ref = ms.reference(m);
      assertCorrectReference(ref, ms.getStoreID(), m);

      MessageReference ref2 = ms.reference(m);

      assertCorrectReference(ref2, ms.getStoreID(), m);
   }

   //////
   ////// Multiple messages
   //////

   public void testRecoverableMessageStore_4() throws Exception
   {
      if (!ms.isRecoverable())
      {
         // we only test recoverable message stores
         return;
      }

      Message[] m = new Message[NUMBER_OF_MESSAGES];
      MessageReference[] refs = new MessageReference[NUMBER_OF_MESSAGES];
      for(int i = 0; i < NUMBER_OF_MESSAGES; i++ )
      {
         m[i] = MessageFactory.createCoreMessage(i, true, 700 + i, 800 + i,
                                             (byte)(i % 10), headers, "payload" + i);

         // recoverable store, non-reliable message, one message
         refs[i] = ms.reference(m[i]);
         assertCorrectReference(refs[i], ms.getStoreID(), m[i]);
      }

      // get reachable reference

      for (int i = 0; i < NUMBER_OF_MESSAGES; i++)
      {         
         refs[i].releaseMemoryReference();
      }

      // there are no strong references to the message reference anymore, so the message store
      // should be cleaned of everything that pertains to those message

      for (int i = 0; i < NUMBER_OF_MESSAGES; i++)
      {
         assertNull(ms.reference(m[i].getMessageID()));
      }
   }
   
   public void testRecoverableMessageStore_6() throws Exception
   {
      if (!ms.isRecoverable())
      {
         // we only test recoverable message stores
         return;
      }

      Message m = MessageFactory.
      createCoreMessage(0, true, 777l, 888l, (byte)9, headers, "payload");

      // non-recoverable store, non-reliable message, one message
      MessageReference ref = ms.reference(m);
      assertCorrectReference(ref, ms.getStoreID(), m);
      
      //Create another reference from that reference
      MessageReference ref2 = ms.reference(m);
          
      ref.releaseMemoryReference();
      
      MessageReference ref3 = ms.reference(m.getMessageID());
      assertNotNull(ref3);
      ref3.releaseMemoryReference();
      
      assertCorrectReference(ref3, ms.getStoreID(), m);
      
      ref2.releaseMemoryReference();

      // there's no strong reference to the unique message reference anymore, so the message store
      // should be cleaned of everything that pertains to that message

      ref3 = ms.reference(m.getMessageID());
      assertNull(ref3);
   }
   
   public void testRecoverableMessageStore_7() throws Exception
   {
      if (!ms.isRecoverable())
      {
         // we only test recoverable message stores
         return;
      }

      Message m = MessageFactory.
      createCoreMessage(0, true, 777l, 888l, (byte)9, headers, "payload");

      // non-recoverable store, non-reliable message, one message
      MessageReference ref = ms.reference(m);
      assertCorrectReference(ref, ms.getStoreID(), m);
      
      MessageReference ref2 = ms.reference(ref.getMessageID());
      assertNotNull(ref2);      
      assertCorrectReference(ref2, ms.getStoreID(), m);
      
      MessageReference ref3 = ms.reference(ref.getMessageID());
      assertNotNull(ref3);      
      assertCorrectReference(ref3, ms.getStoreID(), m);
      
      ref.releaseMemoryReference();
      ref2.releaseMemoryReference();
      ref3.releaseMemoryReference();
      
      MessageReference ref4 = ms.reference(ref.getMessageID());
            
      assertNull(ref4);                
   }

   // Package protected ---------------------------------------------

   // Protected -----------------------------------------------------

   // Private -------------------------------------------------------

   // Inner classes -------------------------------------------------

}
